//
// Copyright 2009 Konstantinos V. Katsaros
//                              ntinos@aueb.gr
//                              http://mm.aueb.gr/~katsaros
//

// This file is part of BitTorrent Implementation for OMNeT++.

//    BitTorrent Implementation for OMNeT++ is free software: you can redistribute
//	  it and/or modify it under the terms of the GNU General Public License as
//    published by the Free Software Foundation, either version 2 of the License,
//	  or (at your option) any later version.

//    BitTorrent Implementation for OMNeT++ is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

//    You should have received a copy of the GNU General Public License
//    along with BitTorrent Implementation for OMNeT++.
//    If not, see <http://www.gnu.org/licenses/>.


/** *********************************************************************** 
 *  STEP 1
 *
 *  Since out Tracker is dynamically deployed at a random point in the 
 *  network we have to know its address so that newly created BitTorrent
 *  peers can learn its address during creation. This means that we do 
 *  not implement a networking procedure for the retrieval of this piece of
 *  information.
 *
 *  Add the following variable in IPv4UnderlayConfigurator.h 
 *
 **************************************************************************/

TransportAddress* trackerAddress;

/** *********************************************************************** 
 *  STEP 2
 *
 *  Add the following method to IPv4UnderlayConfigurator.cc 
 * (do not forget to also add its declaration in the header file).
 *
 **************************************************************************/


/**
 * This is the function that creates and connects the tracker to a randomly selected access router so 
 * that it is reachable by peers. (Derived by IPv4UnderlayConfigurator::createNode())
 *
 */
TransportAddress* IPv4UnderlayConfigurator::createBTNode(const char* type,bool initialize)
{
    const char* device;
    char buf[80];
    
    cModuleType* moduleType = findModuleType(type);
    cModule*  node = moduleType->create(type, parentModule());

    sprintf(buf, type);

    if (!strcmp(type,"Tracker"))
        device  = "i=device/server_l";
    else if (!strcmp(type,"BTHostSeeder"))
        device  = "i=device/mainframe_l";

    node->setGateSize("in", 1);
    node->setGateSize("out", 1);
    node->setDisplayString(device);
    node->buildInside();
    node->scheduleStart(simulation.simTime());

    // create meta information
    IPv4Info* info = new IPv4Info(0, node->id());
    AccessNet* accessNet= check_and_cast<AccessNet*>
    (accessNode[intuniform(0, accessNode.size() - 1)]
                ->submodule("accessNet"));
    info->setAccessNetModule(accessNet);
    // add node to a randomly chosen access net
    info->setNodeID(accessNet->addOverlayNode(node));

    // append index to module name
    node->setName(buf);

    //add node to bootstrap oracle
    bootstrapOracle->addPeer(IPAddressResolver().addressOf(node), info);

    // if the node was not created during startup we have to
    // finish the initialization process manually
    if ( !initialize) {
        for (int i = MAX_STAGE_UNDERLAY + 1; i < NUM_STAGES_ALL; i++) {
            node->callInitialize(i);
        }
    }

    IPvXAddress addr = IPAddressResolver().addressOf(node);
    TransportAddress *address = new TransportAddress(addr);
    address->setPort(80);

    if (!strcmp(type,"Tracker"))
    {
	//Ntinos Katsaros: 03/07/2008
	//Set the address and port at which the tracker listens
	std::string addrString;
	if (!addr.isIPv6())
		addrString = addr.get4().str();
	else
		addrString = addr.get6().str();

	//Now that we have the assinged IP addess (and port) we can change the values of the
	//module parameters. 
	node->submodule("tcpApp",0)->par("address").setStringValue(addrString.c_str());
	node->submodule("tcpApp",0)->par("port") = address->getPort();
    }

    // update display
    setDisplayString();

    return address;
}

/** *********************************************************************** 
 *  STEP 3
 *
 *  Add the following method calls at the the end of the following method: 
 *  void IPv4UnderlayConfigurator::initializeUnderlay(int stage)
 *
 **************************************************************************/

/**
 *  Create and connect the tracker and the initial seeder to a randomly chosen access router. 
 *  Add this code at the end of the following method: 
 *
 *  void IPv4UnderlayConfigurator::initializeUnderlay(int stage)
 *
 *  This will also store the address of the Tracker as explained in STEP 1.
 */

const char* hostType = parentModule()->par("overlayTerminalType");

if (!strcmp(hostType,"BTHost"))
{
	trackerAddress = createBTNode("Tracker",false);
	createBTNode("BTHostSeeder",false);
}



/** *********************************************************************** 
 *  STEP 4
 *
 *  During BitTorrent peer creation, avoid trying to set the following 
 *  overlay related paramaters since BitTorrent  nodes do not participate
 *  in the overlay:
 *
 *  "overlayType", "tier1Type", "tier2Type", "tier3Type" 
 *
 *  This can be achieved by surrounding the node->par(paramName)setStringValue(...)
 *  calls with a check, such as in the following example.
 *
 **************************************************************************/
//====================================================================================================
TransportAddress* IPv4UnderlayConfigurator::createNode(NodeType type, bool initialize)
{
	const char* terminalType = parentModule()->par("overlayTerminalType");
	...

	if (!strcmp(terminalType,"OverlayHost"))
	{
		node->par("overlayType").setStringValue(type.overlayType.c_str());
		node->par("tier1Type").setStringValue(type.tier1Type.c_str());
		node->par("tier2Type").setStringValue(type.tier2Type.c_str());
		node->par("tier3Type").setStringValue(type.tier3Type.c_str());
		node->par("creationOrder").setLongValue((long)numCreated);
		node->setDisplayString("i=device/wifilaptop_l;i2=block/circle_s");
	}
	else
		node->setDisplayString("i=device/wifilaptop_l");

	...

}

/** *********************************************************************** 
 *  STEP 5
 * 
 *  Enable the communication of a BitTorrent client with the Tracker as 
 *  explained in STEP 1.
 *
 *  Add the following code at the end of the createNode() method as shown
 *  below.
 *
 **************************************************************************/

TransportAddress* IPv4UnderlayConfigurator::createNode(NodeType type, bool initialize)
{

	...

	//Get the address of the new node
	IPvXAddress localAddress = IPAddressResolver().addressOf(node);

	//Set the address and port at which the tracker listens
	// and the address and port for this node's tcp client.
	if (!strcmp(terminalType,"BTHost"))
	{
	
		IPvXAddress trackerAddr = trackerAddress->getAddress();
		std::string addrString;
		if (!trackerAddr.isIPv6())
			addrString = trackerAddr.get4().str();
		else
			addrString = trackerAddr.get6().str();
	
		node->submodule("trackerClient")->par("connectAddress") = addrString.c_str();
		node->submodule("trackerClient")->par("connectPort") = trackerAddress->getPort();
	
		if (!localAddress.isIPv6())
			addrString = localAddress.get4().str();
		else
			addrString = localAddress.get6().str();
	
		node->submodule("trackerClient")->par("address") = addrString.c_str();
		node->submodule("trackerClient")->par("port") = address->getPort();

	}

	// update display
	setDisplayString();

	return address;
}
